// f.olofsson2022

#include <Arduino.h>
#include <ESP8266WiFi.h>
#include <WiFiUdp.h>
#include <OSCMessage.h> //https://github.com/CNMAT/OSC
#include "VL53L1X.h"    //https://github.com/pololu/vl53l1x-arduino

//--settings

#define ID 1000                     // EDIT give unique ID to each ESP you program
const char *ssid = "myssid";        // EDIT your accessPoint network name
const char *password = "mypass";    // EDIT your password
const uint16_t in_port = 11082;     // EDIT default input OSC port
const uint16_t out_port = 11083;    // EDIT default output OSC port

// the following defaults can later be set via OSC
#define NUM_MEAN 5      // running mean, more is smoother
#define QUANTISE 2      // quantise by divide and multiply
#define THRESH_MIN 2000 // hysteresis
#define THRESH_MAX 2500 // hysteresis
#define UPDATE_RATE 16  // delay time in milliseconds

//--variables

uint8_t num_mean;
uint8_t quantise;
uint16_t thresh_min;
uint16_t thresh_max;
uint8_t update_rate;

enum mode {
  pinging,
  pausing,
  sending
};
mode state;

IPAddress out_ip;

VL53L1X sensor;

uint8_t last_rstat;
uint16_t last_range;

uint8_t mean_index = 0;
uint16_t mean[256];

WiFiUDP Udp;

void sendOscPing() {
  OSCMessage msg("/f0dist_ping");
  msg.add(ID);
  msg.add(WiFi.localIP()[0]);
  msg.add(WiFi.localIP()[1]);
  msg.add(WiFi.localIP()[2]);
  msg.add(WiFi.localIP()[3]);
  Udp.beginPacket(out_ip, out_port);
  msg.send(Udp);
  Udp.endPacket();
  msg.empty();
}

void setupDistance() {
  Wire.begin(0, 2);
  Wire.setClock(400000);
  sensor.setTimeout(500);
  if (!sensor.init()) {
    while (1);
  }
  sensor.setDistanceMode(VL53L1X::Long);
  sensor.setMeasurementTimingBudget(50000);
  sensor.startContinuous(50);
}

void restart() {
  num_mean = NUM_MEAN;
  quantise = QUANTISE;
  thresh_min = THRESH_MIN;
  thresh_max = THRESH_MAX;
  update_rate = UPDATE_RATE;
  out_ip = WiFi.localIP();
  out_ip[3] = 255; // broadcast
  state = pinging;
}

void oscPong(OSCMessage &msg) {
  setupDistance();
  state = pausing;
}

void oscNumMean(OSCMessage &msg) {
  num_mean = constrain(msg.getInt(0), 0, 255);
}

void oscQuantise(OSCMessage &msg) {
  quantise = msg.getInt(0);
}

void oscThresholds(OSCMessage &msg) {
  thresh_min = msg.getInt(0);
  thresh_max = msg.getInt(1);
}

void oscUpdateRate(OSCMessage &msg) {
  update_rate = msg.getInt(0);
}

void oscRestart(OSCMessage &msg) {
  restart();
}

void setup() {
  delay(10);
  WiFi.mode(WIFI_STA);
  WiFi.begin(ssid, password);
  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
  }
  wifi_set_sleep_type(LIGHT_SLEEP_T); // seems to be the default so no real need
  restart();
  Udp.begin(in_port);
}

void loop() {

  int packet_size = Udp.parsePacket();
  if (packet_size > 0) {
    OSCMessage msg;
    while (packet_size--) {
      msg.fill(Udp.read());
    }
    if (!msg.hasError()) {
      out_ip = Udp.remoteIP();
      msg.dispatch("/pong", oscPong);
      msg.dispatch("/num_mean", oscNumMean);
      msg.dispatch("/quantise", oscQuantise);
      msg.dispatch("/thresholds", oscThresholds);
      msg.dispatch("/update_rate", oscUpdateRate);
      msg.dispatch("/restart", oscRestart);
    }
  }

  if (state == pinging) {
    sendOscPing();
    delay(3000);

  } else {

    sensor.read();

    uint16_t range = sensor.ranging_data.range_mm;
    uint8_t rstat = sensor.ranging_data.range_status;

    mean[mean_index] = range;

    uint32_t sum_range = 0;
    for (uint8_t i = 0; i < num_mean; i++) {
      sum_range += mean[i];
    }
    range = (sum_range / num_mean / quantise) * quantise;

    /*if(state == pausing) {
      if((range != last_range) || (rstat != last_rstat)) {
        state = sending;
      }
    } else {
      state = pausing;
    }*/

    if (state == pausing) {
      if (range < thresh_min) {
        state = sending;
      }
    } else {
      if (range > thresh_max) {
        state = pausing;
      }
    }

    if (state == sending && (range != last_range || rstat != last_rstat)) {
      OSCMessage msg("/f0dist");
      msg.add(ID);
      msg.add(rstat);
      msg.add(range);

      Udp.beginPacket(out_ip, out_port);
      msg.send(Udp);
      Udp.endPacket();
      msg.empty();

      last_rstat = rstat;
      last_range = range;
    }

    mean_index = (mean_index + 1) % num_mean;

    delay(update_rate);
  }
}
